
#include <new>
#include <memory.h>
#include <list>
#include <queue>

#include "common.h"

class Drawable {
protected:
	Drawable() {
		X = Y = Distance = Width = Height = 0;
	}

public:
	int X; // X coord, 0 is left
	int Y; // Y coord, 0 is top
	int Distance; // Z order, smaller is closer
	int Width; // bounding box width in pixels
	int Height; // bounding box height in pixels

	const char *name;
	
	virtual void DrawLine(int CurrentY, Pixel *LineBuffer, int XStart, int XEnd) = 0;
};

class Background : public Drawable {
	Pixel *_Bitmap;
public:
	Background(const char *FileName) {
		_Bitmap = NULL;
		LoadFromBitmap(FileName, &Width, &Height, &_Bitmap);
	}

	virtual void DrawLine(int CurrentY, Pixel *LineBuffer, int XStart, int XEnd) {
		Pixel *p = &_Bitmap[(CurrentY-Y) * Width + (XStart-X)];
		for(int x = XStart; x < XEnd; ++x) {
			LineBuffer[x] = *p;
			++p;
		}
	}
};

class SolidBackground : public Drawable {
	Pixel _Color;

public:
	SolidBackground(Pixel Color) {
		_Color = Color;
	}

	virtual void DrawLine(int CurrentY, Pixel *LineBuffer, int XStart, int XEnd) {
		for(int x = XStart; x < XEnd; ++x)
			LineBuffer[x] = _Color;
	}
};

class CyclingBackground : public Drawable {
	Pixel _Current;
	Pixel _Start;

public:
	CyclingBackground() {
		_Start = 0;
	}

	virtual void DrawLine(int CurrentY, Pixel *LineBuffer, int XStart, int XEnd) {
		if(CurrentY == Y) {
			_Current = _Start;
			++_Start;
		}
		for(int x = XStart; x < XEnd; ++x)
			LineBuffer[x] = _Current++;
	}
};

class Sprite : public Drawable {
	Pixel *_CompressedBitmap;

	struct SpriteSpan {
		int x,y;
		int len;
	};

public:
	Sprite(const char *FileName) {
		Pixel *bm;
		LoadFromBitmap(FileName, &Width, &Height, &bm);

		Pixel *p = bm;
		Pixel background = *p; // topleft pixel
		int total_pixel_count = 0;
		std::list<SpriteSpan> span_list;
		for(int y=0; y<Height; ++y) {
			bool was_in_background = true;
			int span_len = 0;
			int span_start = 0;
			for(int x=0; x<Width; ++x) {
				bool in_background = (*p == background);
				if(was_in_background != in_background) {
					if(!was_in_background && span_len > 0) {
						SpriteSpan sp;
						sp.x = span_start;
						sp.y = y;
						sp.len = span_len;
						span_list.push_back(sp);
						total_pixel_count += sp.len;
					}
					span_len = 1;
					span_start = x;
				} else {
					++span_len;
				}
				was_in_background = in_background;
				++p;
			}
			if(!was_in_background && span_len > 0) {
				SpriteSpan sp;
				sp.x = span_start;
				sp.y = y;
				sp.len = span_len;
				span_list.push_back(sp);
				total_pixel_count += sp.len;
			}
		}

		total_pixel_count += (int)span_list.size(); // empty area sizes in between islands
		total_pixel_count += (int)span_list.size(); // island sizes
		total_pixel_count += Height; // starting pointers for every y
		total_pixel_count += Height; // terminating empty sizes (0xffff)
		_CompressedBitmap = new Pixel[total_pixel_count];
		int store = Height;
		for(int y=0; y<Height; ++y) {
			_CompressedBitmap[y] = store; // starts here
			int empty_start_x = 0;
			while(!span_list.empty() && span_list.front().y == y) {
				int empty_len = span_list.front().x - empty_start_x; // store even if == 0
				_CompressedBitmap[store++] = (Pixel)empty_len;
				SpriteSpan sp = span_list.front();
				span_list.pop_front();
				_CompressedBitmap[store++] = (Pixel)sp.len;
				int offset = sp.y*Width+sp.x;
				for(int i=0; i<sp.len; ++i)
					_CompressedBitmap[store++] = bm[offset+i];
			}
			_CompressedBitmap[store++] = 0xffff;
		}
		assert(store == total_pixel_count);
	}

	virtual void DrawLine(int CurrentY, Pixel *LineBuffer, int XStart, int XEnd) {
		int offset = _CompressedBitmap[CurrentY-Y];
		int span_x = X;
		while(true) {
			span_x += _CompressedBitmap[offset++]; // 0xffff will increment over XEnd
			if(span_x >= XEnd)
				return;
			int span_len = _CompressedBitmap[offset++]; // span start at x coord
			if(span_x+span_len >= XStart) {
				int in_span_offset = XStart-span_x < 0 ? 0 : XStart-span_x;
				int usable_span_len = XEnd-span_x < span_len-in_span_offset ? XEnd-span_x : span_len-in_span_offset;
				Pixel *dst = &LineBuffer[span_x+in_span_offset];
				for(int i = 0; i<usable_span_len; ++i)
					dst[i] = _CompressedBitmap[offset+in_span_offset+i];
			}
			offset += span_len;
		}
	}
};

bool operator < (const Drawable &A, const Drawable &B) {
	debugout("cmp: %s %s\n", A.name, B.name);
	if(A.Distance < B.Distance)
		return true; // first draw farther away
	if(A.Distance > B.Distance)
		return false;
	return (A.X > B.X); // with the same distance draw left to right
}

typedef std::priority_queue<Drawable*> DrawQueue;
std::list<Drawable*> *draw_list = NULL;

Sprite *ship = NULL;
int ship_move = -1;

void InitDemoData() {
	if(draw_list != NULL)
		return;
	draw_list = new std::list<Drawable*>();
	Background *bg;
	bg = new Background("test_png1.bmp");
	bg->X = 20;
	bg->Y = -5;
	bg->Distance = -1;
	bg->name =  "little";
	draw_list->push_back(bg);
	bg = new Background("test_png2.bmp");
	bg->X = 10;
	bg->Y = 10;
	bg->Distance = 0;
	bg->name =  "bigtop";
	draw_list->push_back(bg);
	bg = new Background("test_png2.bmp");
	bg->X = -40;
	bg->Y = 30;
	bg->Distance = 0;
	bg->name =  "bigbott";
	draw_list->push_back(bg);
	SolidBackground *sbg = new SolidBackground(0x001f);
	sbg->Width = 1000;
	sbg->Height = 1000;
	sbg->Distance = 1000;
	sbg->name =  "back";
	draw_list->push_back(sbg);
	ship = new Sprite("aniufo04.bmp");
	ship->X = 10;
	ship->Y = 100;
	ship->Distance = -5;
	ship->name =  "ship";
	draw_list->push_back(ship);
}

std::vector<DrawQueue> draw_list_by_y;

void DrawFrame(int Width, int Height, Pixel *Buffer) {
	InitDemoData();

	if(draw_list_by_y.size() != Height)
		draw_list_by_y.resize(Height);

	// collect drawables grouped by y starting coordinates
	for(std::list<Drawable*>::iterator it = draw_list->begin(); it != draw_list->end(); ++it) {
		Drawable *d = *it;
		if(d->Y+d->Height > 0 && d->Y < Height-1) { // at least 1 line on screen
			int y_index = d->Y < 0 ? 0 : d->Y; // if outside to the top then starts at 0. line
			draw_list_by_y[y_index].push(d);
		}
	}

	std::list<Drawable*> currently_active;

	Pixel *line_start = Buffer;
	for(int y = 0; y < Height; ++y) {
		DrawQueue &starts_this_y = draw_list_by_y[y];
		//if(!starts_this_y.empty()) {
			std::list<Drawable*>::iterator it = currently_active.begin();
			while(true) {
				bool has_list = false;
				while(it != currently_active.end()) {
					Drawable *d = (*it);
					if(y >= d->Y+d->Height) {
						it = currently_active.erase(it);
					} else {
						has_list = true;
						break;
					}
				}
				bool has_pqueue = !starts_this_y.empty();
				if(!has_list && !has_pqueue)
					break;
				bool pop_from_queue;
				if(has_list && has_pqueue) {
					pop_from_queue = (**it < *starts_this_y.top()); // overloaded <
				} else {
					pop_from_queue = has_pqueue;
				}
				if(pop_from_queue) {
					it = currently_active.insert(it, starts_this_y.top());
					debugout("pop name=%s\n", y, starts_this_y.top()->name);
					starts_this_y.pop();
				}
					Drawable *d = (*it);
					int x_start = d->X < 0 ? 0 : d->X;
					int x_end = d->X+d->Width;
					x_end = x_end > Width ? Width : x_end;
					d->DrawLine(y, line_start, x_start, x_end);
					debugout("draw y=%d name=%s\n", y, d->name);
					++it;
				
			}
		//} else {
		//	std::list<Drawable*>::iterator it = currently_active.begin();
		//	while(it != currently_active.end()) {
		//		Drawable *d = (*it);
		//		if(y >= d->Y+d->Height) {
		//			it = currently_active.erase(it);
		//		} else {
		//			d->DrawLine(y, line_start);
		//			++it;
		//		}
		//	}
		//}
		line_start += Width;
	}

#if _DEBUG
	for(int y = 0; y < Height; ++y)
		assert(draw_list_by_y[y].empty());
#endif
}

void SetKeyState(const unsigned char *CurrentState) {
	if((CurrentState[KeyCodeLeft] & KeyStateDown) != 0)
		--ship->X;
	if((CurrentState[KeyCodeRight] & KeyStateDown) != 0)
		++ship->X;
	if((CurrentState[KeyCodeUp] & KeyStateDown) != 0)
		--ship->Y;
	if((CurrentState[KeyCodeDown] & KeyStateDown) != 0)
		++ship->Y;
}
